/*******************************************************************
*
*       Simulated Maximum Likelihood estimation of the
*       stochastic volatiliy model
*       version 1.2
*       February, 1996
*
*       Jon Danielsson (c)
*       Department of Economics
*       University of Iceland
*       101 Reykjavik
*       Iceland
*       e-mail: jond@hag.hi.is
*       ftp  hag.hi.is  
*       WWW  http://www.hag.hi.is/~jond
*       
*       please cite in published work and inform the author.
*
******************************************************************/

#include "svm.h"


extern sml1 sml;
extern double smaxf1(double *x) ;

int last_err;
const int io1 = ios::fixed|ios::right;

double best=-1e100;
int count1;
int sml1::optimize()
{
	count1 = 0;
	int i,t;
	cout.precision(3);
	cout.flags(io1);
	if(Print>2)
		cout<<"\nIteration  omega   delta   nu\t\tlog llik\tbest"<<endl;
//	This sets up and calls my optimization class 

//	am I doing autoscale?
	if(Ascale==YES){
		opt resa(3);			// allocates class, # par=3
		resa.hessian	= NO;		
		resa.tol 	= P;	
		resa.gradient	= NO ;		// I don't supply gradients
		resa.func 	= smaxf1;	// object junction
		resa.par[1] = omega; resa.par[2] = delta; resa.par[3] = nu;
		double vy =var(raw_data,T);
		final_opt=NO;
		for(i=1;i<=3;i++){	
			count1=0;
			best = -1e19;
			scale = exp(0.5*SQR(resa.par[3])/(1.0-SQR(resa.par[2])));
			scale = sqrt(scale/vy);
			for(t=1;t<=T;t++) data[t] = scale*raw_data[t];
			if(Print>3) cout << "\nscale = " << scale << endl;
			resa.par[1]=0.0;
//		cout << "\n\noOmega iteration no " << i <<endl;
//			cout << " var raw " << vy << " var " << var(data,T)<< " scale " << scale << endl;
			scale = -log(scale*scale)*(1.0-resa.par[2]);
			resa.optimize();		
			if(Print>1){
				cout << "Parameters from autoscale iteration no: " <<i<< ":  " ;
				cout <<setw(8) << resa.par[1] << "\t" ;
				cout <<setw(8) << resa.par[2] << "\t" ;
				cout <<setw(8) << resa.par[3] <<endl;
				for(t=1;t<=T;t++) data[t] = raw_data[t];
				omega=resa.par[1];delta=resa.par[2];nu=resa.par[3];
				cout << "log lik. estimate " << func() << endl;
			}
		}
		omega=resa.par[1]; delta = resa.par[2]; nu = resa.par[3];
		scale = exp(0.5*SQR(resa.par[3])/(1.0-SQR(resa.par[2])));
		scale = sqrt(scale/vy);
		for(t=1;t<=T;t++) data[t] = scale*raw_data[t];
		scale = -log(scale*scale)*(1.0-resa.par[2]);
	}	// end autoscale
	opt res(3);			// allocates class, # par=3
	res.hessian	= H;		// estimate hessian
	res.tol 	= P;		// tolerance
	res.gradient	= NO ;		// I don't supply gradients
	res.func 	= smaxf1;	// object junction
	res.par[1] = omega; 
	res.par[2] = delta;
	res.par[3] = nu;

	res.optimize();			// run optimization	
	for(t=1;t<=T;t++) data[t] = raw_data[t];
	omega=res.par[1];delta=res.par[2];nu=res.par[3];
	maxvalue= func(); 		// get max log likelihood

					// get parameters

	if(H==YES){			// if estimated hessian, get s.e.'s
		Hess_omega=res.se[1];
		Hess_delta = res.se[2];
		Hess_nu = res.se[3];
	}
	int iter = count1;

//** If there was problem inthe last 5 iterations, then there MAY be
//   a problem with the results
	if(iter < (last_err+5)) {
		cout << endl << "Possible trouble, iteration of last exeption";
		cout << "\nis " << last_err << " out of " << iter << endl;
		cout << "Suggest rerunning with printed iteration to diagnose";
		cout << endl;
	}
	footer();
	return 0;
}
// The optimizer can't call class funcitons, hence this ineligant solution!
double smaxf1(double *x) {
	double rx;
	rx = sml.OptFunc(x);
	return rx;

}

//	This is the first step towards getting the likelihood
//	checks for inside parameter space, and prints

double sml1::OptFunc(double *x) {
	char ch;
	omega = x[1];
	delta = x[2];
	nu = x[3];
// are the parameters leagal, positive variance etc.
	if(fabs(delta) > 0.999|| nu<0.001) { 
		if(Print>3) {
			cout << "\t" <<setw(8) << x[1] <<setw(8)<<x[2]<<setw(8)<<x[3]<<"\t";
			cout << " Parameters outside of boundaries" << endl;
		}
		return 1e5+delta*delta+omega*omega+nu*nu;
	}

	if(Ascale==YES) omega = omega-scale;
	double res = func();	// get lik estimate

	if(error_code != OK) {
		if(Print>3) {
			cout << "\t" <<setw(8) << x[1] <<setw(8)<<x[2]<<setw(8)<<x[3]<<"\t";
			cout << " Problem with parameters no. "<<error_code << endl;
		}
		return 1e5+delta*delta+omega*omega+nu*nu;
	}else{
		count1++;
		if(Print>2) cout << count1 << "\t" 
			<<setw(8) << x[1] <<setw(8)<<x[2]<<setw(8)<<x[3]<<"\t";
		ch = ' '; if(best< res) {best = res;ch = '*';}
		if(Print>2) cout << setw(10) << res << "\t" <<ch<<endl;
		flush(cout);
	}
		return -res;
}


void sml1::footer(void)
{
if(Print > 0) {
	cout << endl;
	cout << "Function Value = " << maxvalue<<endl ;
	cout << "Running Time   = " << elapsed();
	cout << endl;
	if(H==YES){
		cout << "Parameter Value  s.e."<<endl;
		cout <<"Omega  " << setw(8) << omega<< "(" <<setw(5)<<Hess_omega<<")"<<endl;
		cout <<"Delta  " << setw(8) << delta<< "(" <<setw(5)<<Hess_delta<<")"<<endl;
		cout <<"Nu     " << setw(8) << nu<< "(" <<setw(5)<<Hess_nu<<")"<<endl;
	}else{
		cout <<"Parameter  Value"<<endl;
		cout <<"Omega     " << setw(8) << omega<<endl;
		cout <<"Delta     " << setw(8) << delta<<endl;
		cout <<"Nu        " << setw(8) << nu<<endl;
	}
	cout << endl;
cout << 
"__________________________________________________________________________" ;
	cout << endl;
}
if(Print==0) cout << "log-likelihood " << maxvalue<<endl;
}

