/*******************************************************************
*
*       Simulated Maximum Likelihood estimation of the
*       stochastic volatiliy model
*       version 1.2
*       February, 1996
*
*       Jon Danielsson (c)
*       Department of Economics
*       University of Iceland
*       101 Reykjavik
*       Iceland
*       e-mail: jond@hag.hi.is
*       ftp  hag.hi.is  
*       WWW  http://www.hag.hi.is/~jond
*       
*       please cite in published work and inform the author.
*
******************************************************************/


#include <iostream.h>
#include <iomanip.h>
#include <math.h>
#include <fstream.h>
#include <stdlib.h>
#include "optimize.h"

enum {YES,NO};
static double (*snrfunc)(double *);
double *ppar;
double *simu,*siml;
int igrad;

opt::opt(int x) {
	int i;
	dim = x;
	par = new double [dim+1];
	se = new double [dim+1];
	for(i=1;i<=dim;i++)se[i] = 0.0;
	hessian=NO;
	gradient=NO;
}
opt::~opt() { 
	delete [] par ;
	delete [] se  ;
}



void opt::optimize(void)
{
	ppar = new double [dim+1];
#ifdef SIMPLEX
	nrc_simplex(); 
#elif defined(NAG_OPT)
	nag_simplex();
	if(hessian==YES )
		nag_hessian();
#else
	err("invalid optimizer specified");
#endif
}

void    hess(int k, double *x, double **y,  double delta, double (*funk)(double par[]));

opt::nrc_simplex()
{

	int i,nfunc,j;
	double *x,*y,**p;
	int ndim=3;
	x=new double[1,3];
	y=new double[1,4];
	p=matrix(1,4,1,3);
	for(i=1;i<=ndim;i++) y[i] = par[i] ;
	double lam[4];
	lam[1] = 0.1;
	lam[2] = 0.025;
	lam[3] = 0.2;
	for (i=1;i<=4;i++) {
		for (j=1;j<=3;j++)
			x[j]=p[i][j]=(i == (j+1) ? lam[j]+par[j] : par[j]);
		y[i]=func(x);
	}

	Simplex(p,y,ndim,tol,func,&nfunc);
	for(i=1;i<=ndim;i++) par[i] = p[1][i] ;
	for (i=1;i<=4;i++) {
		for (j=1;j<=3;j++)
			x[j]=p[i][j]=(i == (j+1) ? 0.01+par[j] : par[j]);
		y[i]=func(x);
	}
	Simplex(p,y,ndim,tol/10.0,func,&nfunc);

	for(i=1;i<=ndim;i++) par[i] = p[1][i] ;

//	value = fret1;

//	free_matrix(p,1,4,1,3);
//	delete [] y;
//	delete [] x;
	for (j=1;j<=3;j++)
			y[j] = p[1][j] ;
	if(hessian==YES ) {

		double **hessm;
		hessm  = matrix(1,ndim,1,ndim); 
		hess(3,y,hessm,0.01,func);
		double s11 = hessm[1][1];
		double s12 = hessm[1][2];
		double s13 = hessm[1][3];

		double s22 = hessm[2][2];
		double s23 = hessm[2][3];

		double s33 = hessm[3][3];

		double det = s11*s22*s33;
		det -= s11*s23*s23;
		det -= s12*s12*s33;
		det += 2.0*s12*s13*s23;
		det -= s13*s13*s22;

		se[1] = (s22*s33-s23*s23)/det;
		se[2] = (s11*s33-s13*s13)/det;
		se[3] = (s11*s22-s12*s12)/det;

		for(i=1;i<=3;i++) se[i] = (se[i] < 0.0) ? -999 : sqrt(se[i]);
	}	
	return NULL;
}
opt::nag_simplex()
{
#ifdef NAG_OPT
	int ndim,i,n;

	double *nx,   *w1, *w2, *w3, *w4, *w5, *w6;
	long int iw, maxcal, ifail,ln;
	double fret1;
	ndim = dim;
	snrfunc = func;
	double ftol = tol;

	n = ndim;
	ln = 3;
	iw = 4;
	maxcal = 100;
	ftol = 0.1;

	nx = new double[3];
	w1 = new double[3];	
	w2 = new double[3];	
	w3 = new double[3];	
	w4 = new double[3];	
	w5 = new double[4];	
	w6 = new double[12];	
	ifail = 1;
	for(i=1;i<=ndim;i++) nx[i-1] = par[i] ;
	e04ccf(&ln,nx,&fret1,&ftol,&iw,w1,w2,w3,w4,w5,
			w6,nag_func,monit,&maxcal,&ifail);

	for(i=1;i<=ndim;i++) par[i] = nx[i-1] ;
	value = fret1;
	delete[] w6;
	delete[] w5;
	delete[] w4;
	delete[] w3;
	delete[] w3;
	delete[] w1;
	delete[] nx;
#endif
	return NULL;
}



#define DEBUG
/******************************
*
*	epsrf	relative error, default in <=0;
*
*******************************************************/
opt::nag_hessian()
{
#ifdef NAG_OPT
	int n,nn, lhes, lwork,i;
	int  ifail, iwarn,  mode, msglvl;
	double *hcntrl,*hesian,*sigma,*hforw, *objgrd;
	double *user,*work,*xbb,objf,epsrf;
			
	int info[2], iuser[2];
	nn = dim;

	double ftol = tol;
	snrfunc = func;
	epsrf = -1.0;
	lhes 	= nn+1;
	n	= nn;
	hcntrl	= new double[n+1];
	hesian 	= new double[(n+1)*n+1];
	sigma 	= new double[n*n+1];
	hforw	= new double[n+1];
	objgrd	= new double[n+1];
	user	= new double[n+1];
	work	= new double[n*(n+1)+1];
	xbb	= new double[n+1]; 

	lwork = nn*nn+nn;
	msglvl = 0;
	for(i=1;i<=nn;i++) 
	{
		xbb[i-1] =  par[i];
		hforw[i-1]= -1.0;
	}
	if(gradient==YES) 	mode = 1;
	else			mode = 2;
	ifail = -1;
	if(gradient==YES) igrad=YES; else igrad=NO;
	e04xaf(&msglvl,&n,&epsrf,xbb,&mode,objfun,&lhes,hforw,&objf,
		objgrd,hcntrl, hesian,&iwarn,work,iuser,user,info,&ifail);

	double s11 = hesian[0];
	double s12 = hesian[1];
	double s13 = hesian[2];

	double s22 = hesian[5];
	double s23 = hesian[6];

	double s33 = hesian[10];
	double det = s11*s22*s33;
	det -= s11*s23*s23;
	det -= s12*s12*s33;
	det += 2.0*s12*s13*s23;
	det -= s13*s13*s22;

	se[1] = sqrt((s22*s33-s23*s23)/det);
	se[2] = sqrt((s11*s33-s13*s13)/det);
	se[3] = sqrt((s11*s22-s12*s12)/det);

	delete[] xbb;
	delete[] work;
	delete[] user;
	delete[] objgrd;
	delete[] hforw;
	delete[] sigma;
	delete[] hesian;
	delete[] hcntrl;
#endif
	return 0;
}

// Obective function for simplex
extern int nag_func(int *n, double *x, double *xc)
{
	int i;
	for(i=1;i<=*n;i++) ppar[i] = x[i-1];
	double temp = snrfunc(ppar);
	*xc = temp;
	return 0;
}

// simplex seems to need this
int monit(double *fmin, double *fmax,double *sim,int *n,int *is,int *ncall)
{
	return 0;
}
// hessian calls this

int objfun(int *mode, int *n, double *xbb, double *objf, double *objgrd, int *nstate, int *iuser, double *user)
{
	int i;
    --user;
    --iuser;
    --objgrd;
    --xbb;
	double *grad = new double[*n +1+1];
	for(i=1;i<=*n;i++) ppar[i] = xbb[i];
	*objf = snrfunc(ppar);
	return 0;
} 
