/*******************************************************************
*
*       Simulated Maximum Likelihood estimation of the
*       stochastic volatiliy model
*       version 1.2
*       February, 1996
*
*       Jon Danielsson (c)
*       Department of Economics
*       University of Iceland
*       101 Reykjavik
*       Iceland
*       e-mail: jond@hag.hi.is
*       ftp  hag.hi.is  
*       WWW  http://www.hag.hi.is/~jond
*       
*       please cite in published work and inform the author.
*
******************************************************************/


#include "svm.h"
void SingValueDecomp(double **a,int m,int n,double w[4],double **v);
void SingVBKSb(double **u, double w[4], double **v, int m, int n, double b[4], double x[4]);

regress::~regress(){
//cout << "\ndeallocation in reg" << endl;
	delete [] X;
	delete [] Y;

	delete [] bd1;
	delete [] bd2;
	delete [] bd3;

	delete [] lb1;
	delete [] lb2;
	delete [] lb3;

	delete [] b;
	delete [] w;
	delete [] bd;

// I should deallocate uu and vv however that operation crashes my
// program 

}
regress::init(int xt,int xn,double* data,
	double omega,double delta, double nu,
	double *bb1,double *bb2,double* bb3){


	int i;
// Initialization
	if(all != 0){
		if(T != xt || N != xn) err("init error");
	}
	else {
		all++;
		T=xt;
		N=xn;
		X    = new double[(N+2)*3];
		for(i=0;i<N;i++) X[i]=1.0;
		Y    = new double[N+2];

		lb1     = new double[T+2];
		lb2     = new double[T+2];
		lb3     = new double[T+2];

		bd1     = new double[T+2];
		bd2     = new double[T+2];
		bd3     = new double[T+2];
		uu = new double*[N+1];
		for(i=0; i<=N; i++) uu[i] = new double[4];
		vv = new double*[4];
		for(i=0; i<=4; i++) vv[i] = new double[4];
		w = new double[4];
		b = new double[N+1];
	}
	bb=cc=dd=ee=x1y=x2y=x3y=0.0;
	ii=0;

// ************* calculation of initial values, expected value of Q

	int t;
	double z,nu2,mu,v2,x2;

	nu2 	= SQR(nu);

	v2	= nu2 / ( 1.0 - delta*delta);
	mu 	= omega;
	z 	= exp(-mu + 0.5 * v2 );

	for(t=1;t<=T;t++) {
		x2	= SQR(data[t]) ;
		lb1[t] = bb1[t]	= (1.0 + mu + 0.5*(mu*mu -v2)) * z * x2;
		lb2[t] = bb2[t]	= 1.0 -  x2 *z - 1.0 * mu * x2 * z;
		lb3[t] = bb3[t]	= x2 * z;
	}
	return 0;
}
void regress::add(int n,double ly,double yyy){
	double ly2      = ly*ly;
//	these are for manual regress, see below
	if(RegType==NO){

		bb      += ly;
		cc      += ly2;
		dd      += ly2*ly;
		ee      += ly2*ly2;

		x1y     += yyy;
		x2y     += yyy*ly;
		x3y     += yyy*ly2;
	}

	Y[n-1]  = yyy;
	X[ii++] = ly;
	X[ii++] = ly2;
}

void regress::run(int t,int iter,double& diffindex,double *bb1,double *bb2,double* bb3){


	double ix[4][4];
	int i,j;
	double bd[4];
/*
	this does the linear regression manually by inverting the 
	X'X matrix directly.  This is the fastest way but not robust,
	i.e. we may get get almost singular X'X matrix.  The 
	singular value decomposition method below is much slower
	but but works almost always.

*/
	if(RegType==NO){
		aa = (double) N;

		double det1 = - cc*cc*cc + 2.0*bb*cc*dd - aa*dd*dd - bb*bb*ee + aa*cc*ee;
		ix[1][1] = -dd*dd+cc*ee;
		ix[1][2] = cc*dd-bb*ee;
		ix[1][3] = -cc*cc+bb*dd;
		ix[2][1] = cc*dd-bb*ee;
		ix[2][2] = -cc*cc+aa*ee;
		ix[2][3] = bb*cc-aa*dd;
		ix[3][1] = -cc *cc + bb*dd;
		ix[3][2] = bb*cc-aa*dd;
		ix[3][3] = -bb*bb+aa*cc;

		bd[1] = (ix[1][1]* x1y + ix[1][2]* x2y + ix[1][3]* x3y)/det1;
		bd[2] = (ix[2][1]* x1y + ix[2][2]* x2y + ix[2][3]* x3y)/det1;
		bd[3] = (ix[3][1]* x1y + ix[3][2]* x2y + ix[3][3]* x3y)/det1;
	}
/*
	This is an attempt to decide whether to run the singular value 
	decomposition in addition to the manual method above.

	int ind = 0;	

	for(i=1;i<=3;i++) 
		if(isnan(bd[i])!=0 || bd[i] > 1e30 || bd[i] < -1e30)
			greg3 = 99;
	for(i=1;i<=3;i++) 
		if( fabs(bd[i])> 40.0 || fabs(bd[i]) < 1e-6 )  
			ind = 99;

	if(ind!=0 || greg3  != 0) {
		tmp=regress1(t,bd);
		d = SQR(d1-bd[1])+SQR(d2-bd[2])+SQR(d3-bd[3]);
		d /= SQR(bd[1])+SQR(bd[2])+SQR(bd[3]);
		d *= 100.0;
	}
*/
//	This method calles NRC 

	if(RegType==YES){
		i=0;
		for(j=1;j<=N;j++){
			uu[j][1] = 1.0;
			uu[j][2] = X[i++];
			uu[j][3] = X[i++];
			b[j]	= Y[j-1];
		}
		SingValueDecomp(uu,N,3,w,vv);
		wmax=0.0;
		for(j=1;j<=3;j++) if(w[j] > wmax) wmax=w[j];
		wmin = wmax *1e-10;
		for(j=1;j<=3;j++) if(w[j] < wmin) w[j]=0.0;
		SingVBKSb(uu,w,vv,N,3,b,bd);

	}
//	get the change in coefficients
	diffindex += SQR( bb1[t]-bd[1]); 
	diffindex += SQR( bb2[t]-bd[2]); 
	diffindex += SQR( bb3[t]-bd[3]); 

//	for the first few iteration change coefficients gradually
//	less be get unneccessary overflow

	double m = 0.0;
	if(iter==1) m = 0.7;
	if(iter==2) m = 0.5;
	if(iter==3) m = 0.3;
	if(iter==4) m = 0.1;

	bb1[t] =  bd[1] + m*lb1[t]; 
	bb2[t] =  bd[2] + m*lb2[t];
	bb3[t] =  bd[3] + m*lb3[t];

//	reinitialize 
	bb=cc=dd=ee=x1y=x2y=x3y=0.0;
	ii=0;

//	return OK;
}


//	therse routines area part of the Numerical Recipien in C library


static double maxarg1,maxarg2;
static int imaxarg1,imaxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2))
#define IMIN(a,b) (imaxarg1=(a),imaxarg2=(b),(imaxarg1) < (imaxarg2) ?\
	(imaxarg1) : (imaxarg2))
#define SIGN(a,b) ((b) > 0.0 ? fabs(a) : -fabs(a))




double PytH(double a,double b);


void SingValueDecomp(double** a,int m,int n,double* w,double** v)
{
	int flag,i,its,j,jj,k,l,nm;
	double c,f,h,s,x,y,z;
	double anorm,g,scale;
	double* rv1;

	g=scale=s=anorm=0.0;
	rv1= new double[n+1];

	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
					f=s/h;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}
		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				for (j=l;j<=m;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
					for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=FMAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=IMIN(m,n);i>=1;i--) {
		l=i+1;
		g=w[i];
		for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			for (j=l;j<=n;j++) {
				for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
				f=(s/a[i][i])*g;
				for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else for (j=i;j<=m;j++) a[j][i]=0.0;
		++a[i][i];
	}
	for (k=n;k>=1;k--) {
		for (its=1;its<=30;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if ((double)fabs(rv1[l])+anorm == anorm) {
					flag=0;
					break;
				}
				if ((double)fabs(w[nm])+anorm == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					rv1[i]=c*rv1[i];  // ATH
					if ((double)fabs(f)+anorm == anorm) break;
					g=w[i];
					h=PytH(f,g);
					w[i]=h;
					h=1.0/h;
					c=g*h;
					s= -f*h;
					for (j=1;j<=m;j++) {
						y=a[j][nm];
						z=a[j][i];
						a[j][nm]=y*c+z*s;
						a[j][i]=z*c-y*s;
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]= -v[j][k];
				}
				break;
			}
			if (its == 30) err("No convergence in 30 SVDCMP iterations");
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=PytH(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=PytH(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y*=c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=PytH(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f= c*g + s*y ;
				x= c*y - s*g ;
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
	delete [] rv1;
}

#undef SIGN
#undef MAX

double PytH(double a, double b)
{
	double absa,absb;
	absa=fabs(a);
	absb=fabs(b);
	if(absa>absb) return absa*sqrt(1.0+SQR(absb/absa));
	else return (absb==0.0 ? 0.0: absb*sqrt(1.0+SQR(absa/absb)));
}

void SingVBKSb(double** u, double* w, double** v, int m, int n, 
	double* b, double* x)
{
	int jj,j,i;
	double s;
	double* tmp;
	tmp = new double[n+1];
	for(j=1;j<=n;j++) {
		s=0.0;
		if(w[j]) {
			for(i=1;i<=m;i++) s+= u[i][j]*b[i];
			s /= w[j];
		}
		tmp[j]=s;
	}
	for(j=1;j<=n;j++){
		s=0.0;
		for(jj=1;jj<=n;jj++) s+= v[j][jj]*tmp[jj];
		x[j]=s;
	}
	delete [] tmp;
}
