/*******************************************************************
*
*       Simulated Maximum Likelihood estimation of the
*       stochastic volatiliy model
*       version 1.2
*       February, 1996
*
*       Jon Danielsson (c)
*       Department of Economics
*       University of Iceland
*       101 Reykjavik
*       Iceland
*       e-mail: jond@hag.hi.is
*       ftp  hag.hi.is  
*       WWW  http://www.hag.hi.is/~jond
*       
*       please cite in published work and inform the author.
*
******************************************************************/

#include <iostream.h>
#include <iomanip.h>
#include <math.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>



	/******************************************************************
		THIS SUBROUTINE IS TO EVALUATE THE 2ND DERIVATIVES
		(HESSIAN MATRIX) AT GIVEN POINT FOR GIVEN FUNCTION
		FUNK.

		INPUT	K:	NUMBER OF PARAMETERS;
			X:	THE POINT YOU EVALUATE AT
			FUNK:	FUNCTION SUPPLIED BY YOU
			DELTA:	LENGTH OF DISCRETE INTERVAL IN DIFFERENCE
				APPROXIMATION OF DERIVATIVE
		OUTPUT	Y:	CALCULATED HESSIAN MATRIX AT X
	******************************************************************/

void	hess(int k, double *x, double **y,  double delta, double (*func)(double par[])) 
{
	struct	{ double *xx; } arg[5];
	int	i, j, l, n;

	for( n = 1; n <= 4; n++ ) arg[n].xx = new double [k+1];
	for( i = 2; i <= k; i++ )
		for( j=1; j < i; j++ ) {
			for( l = 1; l <= k; l++) for( n = 1; n <= 4; n++)
				arg[n].xx[l] = x[l];
			arg[1].xx[i] += delta;
			arg[1].xx[j] += delta;
			arg[2].xx[i] += delta;
			arg[2].xx[j] -= delta;
			arg[3].xx[i] -= delta;
			arg[3].xx[j] += delta;
			arg[4].xx[i] -= delta;
			arg[4].xx[j] -= delta;
			y[i][j] = ( func(arg[1].xx) - func(arg[2].xx) - func(arg[3].xx) + func(arg[4].xx) ) / ( 4.0*delta*delta );
		}
	for( l = 1; l <= k; l++) for( n = 1; n <= 4; n++ ) arg[n].xx[l] = x[l];
	for( i = 1; i <= k; i++ ) {
		arg[1].xx[i] += ( 2.0*delta);
		arg[2].xx[i] += ( 1.0*delta);
		arg[3].xx[i] -= ( 1.0*delta);
		arg[4].xx[i] -= ( 2.0*delta);
		y[i][i] = ( -func(arg[1].xx) + 16.0*func(arg[2].xx) -
			30.0*func(x) + 16.0*func(arg[3].xx) - func(arg[4].xx) )
				/ ( 12.0*delta*delta );
		for( n = 1; n <= 4; n++ ) arg[n].xx[i] = x[i];
	}
	for( n = 1; n <= 4; n++ ) delete [] arg[n].xx;
}





double **matrix(int r1, int r2, int c1, int c2)
{
	int i;
	double **m;
	r2 += 1;
	c2 += 1;
	/* allocate pointers to rows */
	m= (double **) calloc((r2+1), sizeof(double)) ;
	if (!m) printf("allocation failure 1 in dmatrix()");

	/* allocate rows and set pointers to them */
	for(i=0;i<=r2;i++) {
		m[i]= (double *) calloc((c2+1),sizeof(double));
		if (!m[i]) printf("allocation failure 2 in dmatrix()");
	}
	/* return pointer to array of pointers to rows */
	return m;
}
void free_matrix(double **m, int r1,int r2, int c1, int c2)
{
	int i;
	r2 += 1;
	c2 += 1;
	for(i=r2;i>=0;i--)  if(m[i]) free(m[i]);
	if(m) free(m);
}

//  Simplex optimization


#define NMAX 5000
#define GET_PSUM \
                         for (j=1;j<=ndim;j++){\
                         for (sum=0.0,i=1;i<=mpts;i++) sum += p[i][j];\
                         psum[j]=sum;}

#define SWAP(a,b) {swap=(a);(a)=(b);(b)=swap;}

double SimplexU(double **p, double y[], double psum[], int ndim,
             double (*funk)(double []), int ihi, double fac);
void Simplex(double **p, double y[], int ndim, double ftol,
       double (*funk)(double []), int *nfunk);

void Simplex(double **p, double y[], int ndim, double ftol,double (*funk)(double []), int *nfunk)
{
       int i,ihi,ilo,inhi,j,mpts;
       mpts=ndim+1;
       double rtol,sum,swap,ysave,ytry;

       double *psum=new double[ndim+2];
       *nfunk=0;
       GET_PSUM
       for (;;) {
            ilo=1;
            ihi = y[1]>y[2] ? (inhi=2,1) : (inhi=1,2);
            for (i=1;i<=mpts;i++) {
                if (y[i] <= y[ilo]) ilo=i;
                if (y[i] > y[ihi]){
                    inhi=ihi;
                    ihi=i;
                } else if (y[i] > y[inhi] && i != ihi) inhi=i;
            }
            rtol=2.0*fabs(y[ihi]-y[ilo])/(fabs(y[ihi])+fabs(y[ilo]));
            if (rtol < ftol) { 
                SWAP(y[1],y[ilo])
                for (i=1;i<=ndim;i++) SWAP(p[1][i],p[ilo][i])
                break;
            }
            if (*nfunk >= NMAX) {
		cout << endl<<"IINMAX exceeded"<<endl;
		exit(1);
	   }
            *nfunk += 2;
            ytry=SimplexU(p,y,psum,ndim,funk,ihi,-1.0);
            if (ytry <= y[ilo])
                ytry=SimplexU(p,y,psum,ndim,funk,ihi,2.0);
            else if (ytry >= y[inhi]) {
                ysave=y[ihi];
                ytry=SimplexU(p,y,psum,ndim,funk,ihi,0.5);
                if (ytry >= ysave) {
                    for (i=1;i<=mpts;i++) {
                           if (i != ilo) {
                               for (j=1;j<ndim;j++)
                                  p[i][i]=psum[j]=0.5*(p[i][j]+p[ilo][j]);
                               y[i]=(*funk)(psum);
			   }
		}	   
                       *nfunk += ndim; 
                        GET_PSUM 
		}
                }else --(*nfunk); 
	}
             delete [] psum;
}
double SimplexU(double **p, double y[], double psum[], int ndim,
             double (*funk)(double []), int ihi, double fac)
{
             int j;
             double fac1,fac2,ytry,*ptry;

             ptry=new double [ndim+2];
             fac1=(1.0-fac)/ndim;
             fac2=fac1-fac;
             for (j=1;j<=ndim;j++) ptry[j]=psum[j]*fac1-p[ihi][j]*fac2;
             ytry=(*funk)(ptry);
             if (ytry < y[ihi]){
                 y[ihi]=ytry;
                 for (j=1;j<=ndim;j++){
                     psum[j] += ptry[j]-p[ihi][j];
                     p[ihi][j]=ptry[j];
                 }
             }

             delete [] ptry;
             return ytry;
}


